/*---------------------------------------------------------------*/
/* DDS Function Generator       (C)ChaN, 2006                    */
/*---------------------------------------------------------------*/


#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <string.h>
#include <stdlib.h>
#include "xitoa.h"
#include "tff.h"
#include "diskio.h"


void get_statics (void*);
DWORD freq2omega (DWORD);
short lin2db (WORD rms);
DWORD db2lin (short db);



/*------------------------------*/
/* DDS control                  */

#define WAVETABLE ((short*)0x8000)
register DWORD Angle1 asm("r2");	/* Ch1 angular register */
register DWORD Angle2 asm("r6");	/* Ch2 angular register */
register WORD Wdreg asm("r10");		/* Temporary buffer */
DWORD Omega1, Omega2;				/* Ch1,Ch2 angular speed register */
short Gain;							/* Gain multiplyer for both channels */
DWORD GainFp;						/* Calcurated peak level */
BOOL DdsRun;						/* DDS is running */

struct {						/* Ch1 wave form statics */
	DWORD CrestFactor;			/* +peak/rms */
	DWORD RailFactor;			/* +peak/p-p */
} WfParm;

#define MAX_FREQ	4500000UL	// x10mHz
#define MAX_VOLTPK	7800UL		// Maximum peak voltage in mV
#define P_VOLT		17207		// Gain = ((vpk_fp >> 12) * P_VOLT) >> 16;	// 32767@MAX_VOLTPK
BYTE LevelUnit;					/* 0:Vrms, 1:Vpp, 2:dBV, 3:dBm */
short Level;					/* Current level in display */
BOOL FreqTrack;					/* Ch1-Ch2 freqency tracking mode */
DWORD Freq1, Freq2;				/* Ch1,Ch2 frequency (Omega reg.) */
short Phase2;					/* Ch1-Ch2 phase (valid in tracking mode) */

BYTE CurrWfm;					/* Waveform index currentry loaded */


/*------------------------------*/
/* User I/F control             */

#define K_FWD	_BV(PF0)
#define K_REV	_BV(PF1)
#define K_LEFT	_BV(PF2)
#define K_RIGHT	_BV(PF3)
#define K_UP	_BV(PF4)
#define K_DOWN	_BV(PF5)
#define	K_BLINK	0x81
#define LCD_E	_BV(PE2)
#define LCD_RW	_BV(PE3)
#define LCD_RS	_BV(PE4)
#define ROW_NAME	0
#define ROW_LEVEL	1
#define ROW_FREQ1	2
#define ROW_FREQ2	3
#define ROW_PHASE	4

volatile BYTE KeyCmd;			/* Key command register */
volatile BYTE Beep, Blink;		/* Beep control, Blink control */
BYTE CsrRow, CsrCol;

BYTE CurrItem;						/* 0:waveform, 1:voltage, 2:freq1, 3:freq2, 4:phase2 */
BYTE CurrCol[5] = {0, 1, 0, 0, 0};	/* column position for each item */

prog_uint16_t mag10v[] = {1000, 100, 10, 1};
prog_uint32_t mag10f[] = {100000, 10000, 1000, 100, 10, 1};


/*------------------------------*/
/* MMC/File I/O contorl         */

extern volatile BYTE MmcTimer, MmcStat;
union {
	DWORD	dw;
	WORD	w[2];
} FatTime;
FILINFO finfo;
FATFS fs;					/* File system object */
DIR dir;					/* Directory object */
FIL file1;					/* File object */



/*------------------------------*/
/* Waveform Information         */

#define MAX_WFM 30
int nWaves;
char *NamePtr[MAX_WFM];
char NameTable[800];



/*---------------------------------------------------------*/
/* RTC function (always returns a specified time)          */

DWORD get_fattime ()
{
	return FatTime.dw;
}




/*--------------------------------------------------------------------------*/
/* DDS control                                                              */



static
void dly_ms (WORD ms)
{
	TCCR0 = 0b00000110;
	do {
		TCNT0 = -78;
		TIFR = _BV(TOV0);
		while ((TIFR & _BV(TOV0)) == 0);
	} while (--ms);
}



static
void dds_start ()
{
	PORTD &= ~_BV(PD5);		// Release DAC reset
	EIFR = _BV(INTF0);		// Enable DDS interrupt by LRCK
	EIMSK = _BV(INTF0);
	PORTD &= ~_BV(PD6);		// Release Mute
	DdsRun = TRUE;
}



static
void dds_stop ()
{
	EIMSK = 0;				// Disable DDS interrupt
	PORTD |= _BV(PD6);		// Mute
	*(short*)0x7FFE = 0;
	Wdreg = 0;
	DdsRun = FALSE;
}



static
short dds_setlevel (short level)
{
	DWORD gtmp, *fc;


	if (LevelUnit & 2) {
		gtmp = WfParm.CrestFactor / 100 * db2lin(level);
		if (LevelUnit & 1) {
			if (gtmp >= 16000000)
				gtmp = 50764UL * (gtmp >> 16);
			else
				gtmp = (50764UL * (gtmp >> 8)) >> 8;
		}
		if (gtmp > (MAX_VOLTPK << 16)) {
			gtmp = (MAX_VOLTPK << 16);
			if (LevelUnit & 1)
				gtmp = 84607UL * (gtmp >> 16);
			level = lin2db((short)(gtmp / WfParm.CrestFactor));
		}
	} else {
		fc = (LevelUnit & 1) ? &WfParm.RailFactor : &WfParm.CrestFactor;
		gtmp = (DWORD)level * *fc;
		if (gtmp > (MAX_VOLTPK << 16)) {
			gtmp = (MAX_VOLTPK << 16);
			level = (short)(gtmp / *fc);
		}
	}
	GainFp = gtmp;

	gtmp = ((GainFp >> 12) * P_VOLT) >> 12;
	if (gtmp > 32767) {
		gtmp >>= 4;
		cli();
		Gain = -(short)gtmp;
		sei();
		dly_ms(1);
		PORTD &= ~_BV(7);
		return level;
	}
	PORTD |= _BV(7);
	dly_ms(1);
	cli();
	Gain = -(short)gtmp;
	sei();
	return level;
}



/*--------------------------------------------------------------------------*/
/* LCD controls                                                             */


static
void lcd_write (BYTE rs, BYTE data, BYTE busy)
{
	BYTE d;


	TIMSK = 0;	// Disable 800Hz Timer (key scan)

	PORTF = data;

	if (busy) {	// Busy check
		PORTE &= ~LCD_RS;	// RS=L
		PORTE |=  LCD_RW;	// RW=H
		do {
			PORTE |=  LCD_E;	// E=H
			asm volatile("nop");
			asm volatile("nop");
			asm volatile("nop");
			asm volatile("nop");
			asm volatile("nop");
			d = PINF;
			PORTE &= ~LCD_E;	// E=L
		} while (d & 0x80);		// loop until BF=0
		d = 20;					// Delay 1.5 fosc after BF goes 0
		do {
			asm volatile("nop");
		} while (--d);
	}

	if (rs) PORTE |=  LCD_RS;	// RS=H
	else    PORTE &= ~LCD_RS;	// RS=L
	DDRF = 0xFF;
	PORTE &= ~LCD_RW;	// RW=L
	PORTE |=  LCD_E;	// E=H
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	PORTE &= ~LCD_E;	// E=L
	PORTF = 0xFF;
	DDRF = 0;

	TIMSK = _BV(OCIE2);		// Reenable 800Hz Timer (key scan)
}



static
void lcd_setpos (BYTE row, BYTE col)
{
	CsrRow = row;
	CsrCol = col;
	if (row & 2) col += 20;
	if (row & 1) col += 64;
	lcd_write(0, col | 0x80, 1);
}



static
void lcd_putc (char chr)
{
	BYTE n;


	if (chr == '\f') {
		lcd_write(0, 0x01, 1);
		CsrRow = CsrCol = 0;
		return;
	}
	if (chr == '\n') {
		n = (++CsrRow >= 4) ? 20 : 0;
		lcd_setpos(CsrRow, n);
		return;
	}
	if (chr == '\r') {
		lcd_setpos(CsrRow, 0);
		return;
	}
	if (CsrCol < 20) {
		lcd_write(1, chr, 1);
		CsrCol++;
	}
}



static
void lcd_init ()
{
	dly_ms(30);
	lcd_write(0, 0b00111000, 0);
	dly_ms(10);
	lcd_write(0, 0b00111000, 0);
	dly_ms(3);
	lcd_write(0, 0b00111000, 0);
	lcd_write(0, 0b00111000, 1);	// 2 line, 5x7 font
	lcd_write(0, 0b00000001, 1);	// Clear
	lcd_write(0, 0b00000110, 1);	// Adress increment, No shift
	lcd_write(0, 0b00001101, 1);	// Display ON, Cursor OFF, Blink ON
}



static
void put_freq (BYTE ch, DWORD val)
{
	lcd_setpos(ch + 1, 0);
	xprintf(PSTR("Ch%u %5u.%02uHz"), ch, (WORD)(val / 100), (WORD)(val % 100));
}



static
void put_level (short val)
{
	short v1;


	lcd_setpos(1, 0);
	if (LevelUnit >= 2) {
		xputs(PSTR("Vout  "));
		v1 = abs(val / 100);
		if (v1 < 10) xputc(' ');
		xprintf(PSTR("%c%u.%02u"), (val < 0) ? '-' : '+', v1, abs(val % 100));
		xputs((LevelUnit & 1) ? PSTR("dBm ") : PSTR("dBV "));
	} else {
		xprintf((LevelUnit & 1) ? PSTR("Vout %3d.%03uVp-p") : PSTR("Vout %3d.%03uVrms"),
		val / 1000, abs(val % 1000));
	}
	xputs(PSTR(" EMF"));
}



static
void put_phase (short val)
{
	lcd_setpos(3, 15);
	if (FreqTrack) {
		if (val >= 100) {
			xprintf(PSTR("+%d\xDF"), val);
			return;
		}
		if (val >= 10) {
			xprintf(PSTR(" +%d\xDF"), val);
			return;
		}
		if (val >= 0) {
			xprintf(PSTR("  +%d\xDF"), val);
			return;
		}
		xprintf(PSTR("%4d\xDF"), val);
		return;
	}
	xprintf(PSTR(" ***\xDF"), val);
}



static
void put_name (BYTE index, BYTE wfm)
{
	char c, *p;


	lcd_setpos(0, 0);
	if (index)
		xprintf(PSTR("%02u\x7E"), wfm + 1);
	p = NamePtr[wfm];
	while ((c = *p++) != '=' && c) xputc(c);
	while (CsrCol < 20) xputc(' ');
}



/*--------------------------------------------------------------------------*/
/* File controls                                                            */


static
int load_index ()
{
	BYTE c;
	int i = 0;
	WORD rb;
	FIL fidx;


	nWaves = 0;
	if (!(disk_status() & STA_NODISK)) {
		dly_ms(500);
		if (f_mountdrv() == FR_OK &&
			f_open(&fidx, "index.txt", FA_READ | FA_OPEN_EXISTING) == FR_OK) {
			do {
				NamePtr[nWaves] = &NameTable[i];
				do {
					c = 0;
					if (f_read(&fidx, &c, 1, &rb) || !rb) break;
					if (c == '\n') break;
					if (c < ' ') continue;
					NameTable[i++] = c;
				} while (i < sizeof(NameTable));
				if (c != '\n') break;
				NameTable[i++] = '\0';
			} while (++nWaves < MAX_WFM);
			f_close(&fidx);
		}
	}

	return nWaves;
}



static
BOOL load_waveform (BYTE wfm)
{
	long val;
	char *lp, c;
	const char *fname;
	short *wt;
	int n, ext;
	BYTE buff[512], lbuff[20], nbuff[40];
	WORD li, rs, ri;
	FIL fobj;
	FILINFO fi_csv, fi_bin;


	fname = NamePtr[wfm];	/* Pick-up file name from waveform list */
	do {
		c = *fname++;
		if (!c) return FALSE;
	} while (c != '=');
	ext = strlen(fname);
	if (ext >= sizeof(nbuff) - 5) return FALSE;
	strcpy(nbuff, fname);

	strcpy(&nbuff[ext], ".csv");	/* Get time stamp of csv file */
	if (f_stat(nbuff, &fi_csv)) return FALSE;

	strcpy(&nbuff[ext], ".bin");	/* Load cache file if it is in sync */
	if (f_stat(nbuff, &fi_bin) == FR_OK
		&& fi_csv.fdate == fi_bin.fdate
		&& fi_csv.ftime == fi_bin.ftime
		) {
		if (f_open(&fobj, nbuff, FA_READ | FA_OPEN_EXISTING) == FR_OK) {
			if (   f_read(&fobj, WAVETABLE, 32768, &rs) == FR_OK && rs == 32768	// Load waveform
				&& f_read(&fobj, &WfParm, sizeof(WfParm), &rs) == FR_OK && rs == sizeof(WfParm)	// Load statics
				) {
				f_close(&fobj);
				return TRUE;
			}
			f_close(&fobj);
		}
	}
	/* Cache file could not be loaded */

	strcpy(&nbuff[ext], ".csv");	/* Open csv file */
	if (f_open(&fobj, nbuff, FA_READ | FA_OPEN_EXISTING)) return FALSE;

	rs = ri = sizeof(buff);
	for (wt = WAVETABLE, n = 8192; n; wt++, --n) {
		li = 0;
		do {
			if (li >= sizeof(lbuff)) goto lw_err;
			if (ri >= rs) {
				if (rs != sizeof(buff) || f_read(&fobj, buff, sizeof(buff), &rs) || !rs) goto lw_err;
				ri = 0;
			}
			c = buff[ri++];
			if (c == ',') c = ' ';
			lbuff[li++] = c;
		} while (c != '\n');
		lp = lbuff;
		if (!xatoi(&lp, &val)) goto lw_err;
		*wt = (short)val;
		if (!xatoi(&lp, &val)) goto lw_err;
		*(wt+8192) = (short)val;
	}
	f_close(&fobj);

	get_statics(&WfParm);	/* Get wave form statics */

	strcpy(&nbuff[ext], ".bin");	/* Update cache file */
	FatTime.w[0] = fi_csv.ftime;
	FatTime.w[1] = fi_csv.fdate;
	if (f_open(&fobj, nbuff, FA_WRITE | FA_CREATE_ALWAYS)) return TRUE;
	f_write(&fobj, WAVETABLE, 32768, &rs);
	f_write(&fobj, &WfParm, sizeof(WfParm), &rs);

lw_err:
	f_close(&fobj);
	return n ? FALSE : TRUE;
}




/*--------------------------------------------------------------------------*/
/* I/O initializetion                                                       */


static
void IoInit ()
{
	PORTG = 0b00111111;		// PortG pull-up

	/* Analog Contorl */
	PORTD = 0b01111110; 	// Port D
	DDRD  = 0b11100000;
	EICRA = 0b00000011;		// Enable INT0 in rising edge

	/* User I/F */
	PORTE = 0b11000001;		// Port E
	DDRE  = 0b00111110;
	PORTF = 0b11111111;		// Port F

	/* MMC I/F */
	PORTB = 0b11110101;		/* Enable drivers */
	DDRB  = 0b00000111;
	SPCR = 0b01010000;		/* Initialize SPI port (Mode 0) */
	SPSR = 0b00000001;

	/* External Memory */
	PORTA = 0b11111111;		// Port A pull up
	MCUCR = 0b11000000; 	// Enable external memory with 1ws

	/* Timer */
	OCR2 = 96-1;			// Timer2: 800Hz interval (OC2)
	TCCR2 = 0b00001100;
	TIMSK = _BV(OCIE2);		// Enable TC0.oc interrupt

	/* Beep */
	OCR3A = 2458-1;
	OCR3C = 0;
	TCCR3A = 0b00000100;

	sei();
}




/*-----------------------------------------------------------------------*/
/* Command Processing                                                    */


static
void return_cursor ()
{
	BYTE row, col, n;


	row = CurrItem;
	col = CurrCol[row];
	switch (row) {
	case 0 :
		col = 0;
		break;
	case 1 :
		n = (LevelUnit >= 2) ? 2 : 1;
		if (col >= n) col++;
		col += 7;
		break;
	case 2 :
	case 3 :
		if (col >= 4) col++;
		col += 5;
		break;
	case 4 :
		col += 17;
	}
	if (row > 3) row--;
	lcd_setpos (row, col);
}



static
void proc_wave(int dir)
{
	BYTE n;


	n = CurrCol[0] + dir;
	if (n >= nWaves)
		n = (dir > 0) ? 0 : nWaves - 1;
	CurrCol[0] = n;
	put_name(1, n);
}



static
void proc_level(int dir)
{
	short lv;
	WORD mv;
	BYTE lu;


	if (CurrCol[1] == 4) {
		lu = dir + LevelUnit;
		if (lu > 3)
			lu = (dir > 0) ? 0 : 3;
		LevelUnit = lu;
		if ((lu & 2) == 0) {
			lv = (short)(GainFp / ((lu & 1) ? WfParm.RailFactor : WfParm.CrestFactor));
		} else {
			mv = (WORD)(GainFp / WfParm.CrestFactor);
			if (lu & 1)
				mv = (WORD)((84607UL * mv) >> 16);
			lv = lin2db(mv);
		}
		put_level(Level = lv);
		return;
	}

	lv = Level + dir * pgm_read_word_near(&mag10v[CurrCol[1]]);
	if (LevelUnit >= 2) {	/* In log scale, clip between +20.00 to -60.00 */
		if (lv > 2000) lv = 2000;
		if (lv < -6000) lv = -6000;
	} else {				/* In linear scale, clip between 16.000 to 0.001 */
		if (lv <= 0) return;
		if (lv > 16000) lv = 16000;
	}
	put_level(Level = dds_setlevel(lv));
}



static
void proc_freq1(int dir)
{
	DWORD freq, omega;


	freq = (long)dir * pgm_read_dword(&mag10f[CurrCol[2]]) + Freq1;
	if (freq > 4500000 || !freq) {
		if (dir < 0) return;
		freq = 4500000;
	}
	Freq1 = freq;
	put_freq(1, freq);
	omega = freq2omega(freq);
	if (FreqTrack) {
		Freq2 = freq;
		put_freq(2, freq);
		cli();
		Omega2 = omega;
	}
	cli();
	Omega1 = omega;
	sei();
}



static
void proc_freq2(int dir)
{
	DWORD freq, omega;


	freq = (long)dir * pgm_read_dword(&mag10f[CurrCol[2]]) + Freq2;
	if (freq > 4500000 || !freq) {
		if (dir < 0) return;
		freq = 4500000;
	}
	Freq2 = freq;
	if (FreqTrack) {
		FreqTrack = FALSE;
		Phase2 = 0;
		put_phase(0);
	}
	put_freq(2, freq);
	omega = freq2omega(freq);
	cli();
	Omega2 = omega;
	sei();
}



static
void proc_phase (int dir)
{
	short ph;
	long ag;


	ph = dir * (CurrCol[4] ? 1 : 10) + Phase2;
	if (ph >= 180) ph -= 360;
	if (ph < -180) ph += 360;
	Phase2 = ph;
	ag = 2982616L * ph;
	if (!FreqTrack) {
		FreqTrack = TRUE;
		Freq2 = Freq1;
		put_freq(2, Freq1);
		cli();
		Omega2 = Omega1;
	}
	cli();
	Angle2 = Angle1 + ag;
	sei();
	put_phase(ph);
}



static
void blink_hold ()
{
	Beep = 1;
	Blink = 50;
	lcd_write(0, 0b00001100, 1);
}



static
void proc_cmd (BYTE cmd)
{
	static char div_enc;	// Encoder count divider
	BYTE m = CurrItem;

	if (cmd == K_BLINK) {
		lcd_write(0, 0b00001101, 1);
		return;
	}

	if (cmd == K_FWD) {
		if (++div_enc >= 2) {
			div_enc = 0;
			blink_hold();
			if (m == 0) {
				proc_wave(1); return;
			}
			if (m == 1) {
				proc_level(1); return;
			}
			if (m == 2) {
				proc_freq1(1); return;
			}
			if (m == 3) {
				proc_freq2(1); return;
			}
			if (m == 4) {
				proc_phase(-1); return;
			}
		}
		return;
	}
	if (cmd == K_REV) {
		if (--div_enc <= -2) {
			div_enc = 0;
			blink_hold();
			if (m == 0) {
				proc_wave(-1); return;
			}
			if (m == 1) {
				proc_level(-1); return;
			}
			if (m == 2) {
				proc_freq1(-1); return;
			}
			if (m == 3) {
				proc_freq2(-1); return;
			}
			if (m == 4) {
				proc_phase(1); return;
			}
		}
		return;
	}
	div_enc = 0;
	Beep = 12;

	if (cmd == K_UP) {
		if (CurrItem) {
			CurrItem--;
			if(CurrItem == 0) ;
		}
		return;
	}

	if (cmd == K_DOWN) {
		if (DdsRun && CurrItem < 4) {
			if (!CurrItem)
				put_name(0, CurrCol[0] = CurrWfm);
			CurrItem++;
		}
		return;
	}

	if (cmd == K_LEFT) {
		switch (CurrItem) {
		case 1 :
			if (CurrCol[1]) CurrCol[1]--;
			break;
		case 2 :
		case 3 :
			if (CurrCol[2]) CurrCol[2]--;
			CurrCol[3] = CurrCol[2];
			break;
		case 4 :
			CurrCol[4] = 0;
		}
		return;
	}

	if (cmd == K_RIGHT) {
		switch (CurrItem) {
		case 0 :
			xputs(PSTR("                    \rLoading..."));
			dds_stop();
			if (load_waveform(CurrCol[0])) {
				put_level(Level = dds_setlevel(Level));
				Angle2 = Angle1 + 1491308L * Phase2;
				dds_start();
				put_name(0, CurrWfm = CurrCol[0]);
				break;
			}
			xputs(PSTR("Error!"));
			Beep = 250;
			break;
		case 1 :
			if (CurrCol[1] < 4) CurrCol[1]++;
			break;
		case 2 :
		case 3 :
			if (CurrCol[2] < 5) CurrCol[2]++;
			CurrCol[3] = CurrCol[2];
			break;
		case 4 :
			CurrCol[4] = 1;
		}
		return;
	}
}




/*-----------------------------------------------------------------------*/
/* Main                                                                  */


int main ()
{
	IoInit();					/* Initialize internal peripherls */
	lcd_init();					/* Initialize LCD */
	xfunc_out = lcd_putc;		/* Join xitoa module to lcd module */
	FatFs = &fs;				/* Enable FatFs module */


	Level = 100;				/* Level = 0.1Vrms */
	Freq1 = Freq2 = 100000UL;	/* Freqency = 1000.00Hz */
	Omega1 = Omega2 = freq2omega(Freq1);
	FreqTrack = TRUE;			/* Tracking mode */

	for (;;) {
		xputs(PSTR("\fLoading index..."));
		while (!load_index());
		xprintf(PSTR("\f%u Waveforms"), nWaves);
		Beep = 12;
		dly_ms(1000);

		/* Initial display */
		xputc('\f');
		put_level(Level);
		put_freq(1, Freq1);
		put_freq(2, Freq2);
		put_phase(Phase2);
		put_name(1, CurrCol[0] = 0);
		CurrItem = 0;
		return_cursor();

		/* User command processing loop */
		KeyCmd = 0;
		while (!(disk_status() & STA_NOINIT) || CurrItem) {
			if (!KeyCmd) continue;
			proc_cmd(KeyCmd);
			KeyCmd = 0;
			return_cursor();
		}

		dds_stop();		/* Card removed */
		Beep = 250;
	}
}



/*---------------------------------------------------------*/
/* 800Hz timer interrupt generated by OC2                  */
/*---------------------------------------------------------*/


ISR(TIMER2_COMP_vect)
{
	static BYTE pv, Div8, KeyFil[3];
	BYTE n, s, k;


	sei();

	k = KeyFil[0];						/* Sample contacts */
	KeyFil[0] = s = ~PINF;

	if (s & 2) s ^= 1;					/* Encoder process */
	n = KeyFil[2]; KeyFil[2] = s;
	s = (s - n) & 3;
	if (s == 1) KeyCmd = K_FWD;
	if (s == 3) KeyCmd = K_REV;

	if (k == KeyFil[0]) {				/* Button process */
		n = KeyFil[1]; KeyFil[1] = k;
		n ^= k;
		n &= k;
		n &= 0b00111100;
		if (n) KeyCmd = n;
	}
	if (Beep) {
		Beep--;
		TCCR3B = 0b00001001;
	} else {
		TCCR3B = 0;
	}
	Div8 = (Div8 + 1) & 7;
	if (Div8) return;

	if (!KeyCmd) {						/* Delayed blink restart */
		n = Blink;
		if (n--) {
			Blink = n;
			if (!n) KeyCmd = K_BLINK;
		}
	}

	n = MmcTimer;					/* 100Hz decrement timer for MMC */
	if (n) MmcTimer = --n;

	n = pv;
	pv = PINB & 0x30;		/* Sapmle socket switch */

	if (n == pv) {					/* Have contacts stabled? */
		s = MmcStat;
		if (pv & 0x10)				/* INS = H (Socket empty) */
			s |= (STA_NODISK | STA_NOINIT);
		else						/* INS = L (Card inserted) */
			s &= ~STA_NODISK;
		if (pv & 0x20)				/* WP is H (write protected) */
			s |= STA_PROTECT;
		else						/* WP is L (write enabled) */
			s &= ~STA_PROTECT;
		MmcStat = s;
	}

}
